@@ -13,6 +13,10 @@ from . import (
 )
 
 class DataFrame(abc.SQL):
+    def __init__(self, db, sql):
+        super().__init__(db, sql)
+        self._table = self.sql.subquery()
+    
     def fetch(self):
         with self.db.connect() as con:
             cur = con.execute(self.sql)
@@ -21,11 +25,11 @@ class DataFrame(abc.SQL):
     
     @property
     def table(self):
-        return self.from_
+        return self._table
     
     @property
     def columns(self):
-        return [c.name for c in self.sql.subquery().c]
+        return tuple(c.name for c in self.table.c)
     
     def items(self):
         for k in self.columns:
@@ -43,7 +47,7 @@ class DataFrame(abc.SQL):
             'constraints': c.constraints,
             'autoincrement': c.autoincrement,
             'computed': c.computed,
-        } for c in self.sql.c])
+        } for c in self.table.c])
         return info
         
     def __getitem__(self, key):
@@ -76,11 +80,11 @@ class DataFrame(abc.SQL):
         elif isinstance(col_key, str):
             sql = sa.select(self.table.c[col_key])
         elif isinstance(col_key, collections.abc.Iterable):
-            sql = sa.select(*[self.table.c[k] for k in col_key])
+            sql = sa.select(*self.table.c[tuple(col_key)])
         else:
             raise TypeError(f"col_key must be a str, slice, Ellipsis, or Iterable, but {type(col_key)=}.")
 
-        if isinstance(row_key, Column):
+        if isinstance(row_key, IndexableColumn):
             sql = sql.where(row_key.column)
         elif isinstance(row_key, int):
             start, stop = row_key, row_key + 1
@@ -92,10 +96,14 @@ class DataFrame(abc.SQL):
             stop = -1 if row_key.stop is None else row_key.stop
             sql = sql.slice(start, stop)
         else:
-            raise TypeError(f"row_key must be an Column, int, slice, or Ellipsis, but {type(row_key)=}.")
+            raise TypeError(f"row_key must be an IndexableColumn, int, slice, or Ellipsis, but {type(row_key)=}.")
             
-        from_ = self.table.c[col_key] if result_class is Column else self.table
-        return result_class(self.db, from_, sql)
+        if result_class is Column:
+            if isinstance(row_key, slice) and row_key.start is None and row_key.stop is None:
+                return IndexableColumn(self.db, self.table.c[col_key])
+            return Column(self.db, sql)
+            
+        return result_class(self.db, sql)
     
     def merge(self, df, how='inner', sort=False, on=None, left_on=None, right_on=None, suffixes=('_x', '_y')):
         if how == 'cross':
@@ -107,7 +115,7 @@ class DataFrame(abc.SQL):
             right_cols = [df.table.c[col].label(f'{col}{suffixes[1]}') if col in overlap else df.table.c[col] for col in df.columns]
             sql = sa.select(*left_cols, *right_cols)
             
-            return DataFrame(self.db, sql.subquery(), sql)
+            return DataFrame(self.db, sql)
         
         if on is not None:
             if not (left_on is None and right_on is None):
@@ -154,10 +162,10 @@ class DataFrame(abc.SQL):
             else:
                 sql = sql.order_by(*[left.c[col] for col in left_on], *[right.c[col] for col in right_on if col not in on], left.c.row_number, right.c.row_number)
             
-        return DataFrame(self.db, table, sql)
+        return DataFrame(self.db, sql)
     
     def sort_values(self, by):
-        return DataFrame(self.db, self.table, self.sql.order_by(by))
+        return DataFrame(self.db, self.sql.order_by(by))
     
     def groupby(self, columns):
         return DataFrameGroupBy(self, columns)
@@ -183,8 +191,8 @@ class DataFrame(abc.SQL):
 class DataFrameGroupBy:
     def __init__(self, df, columns):
         self.df = df
-        self.columns = list(columns)
-        self.ungrouped_columns = [col for col in self.df.columns if col not in self.columns]
+        self.columns = tuple(columns)
+        self.ungrouped_columns = tuple(col for col in self.df.columns if col not in self.columns)
     
     def __getitem__(self, key):
         if not isinstance(key, str):
@@ -218,13 +226,13 @@ class DataFrameGroupBy:
             column = func(self.df.table.c[column]).label(k)
             columns.append(column)
         
-        return DataFrame(self.df.db, self.df.table, sa.select(*[self.df.table.c[col] for col in self.columns], *columns).group_by(*self.columns))
+        return DataFrame(self.df.db, sa.select(*self.df.table.c[self.columns], *columns).group_by(*self.columns))
         
     def transform(self, *args, **kwargs):
         raise NotImplementedError()
         
     def nunique(self):
-        return DataFrame(self.df.db, self.df.table, sa.select(*self.df.table.c[tuple(self.columns)], *[sa.func.count(self.df.table.c[col].distinct()).label(col) for col in self.ungrouped_columns]).group_by(self.columns))
+        return DataFrame(self.df.db, sa.select(*self.df.table.c[self.columns], *[sa.func.count(self.df.table.c[col].distinct()).label(col) for col in self.ungrouped_columns]).group_by(*self.columns))
     
     def count(self):
         return self.agg('count')
@@ -249,7 +257,7 @@ class GroupByNthSelector:
             raise KeyError(f"key must be an int or a slice, but {type(key)=}.")
             
         df = sa.select(self.gb.df.table, sa.func.row_number().over(partition_by=self.gb.columns).label('row_number'), sa.func.row_number().over().label('order')).subquery()
-        return DataFrame(self.gb.df.db, self.gb.df.table, sa.select(*[df.c[col] for col in self.gb.df.columns]).where(df.c.row_number.between(start, stop)).order_by(df.c.order))
+        return DataFrame(self.gb.df.db, sa.select(*df.c[self.gb.df.columns]).where(df.c.row_number.between(start, stop)).order_by(df.c.order))
     
     
 class ColumnGroupBy:
@@ -258,7 +266,7 @@ class ColumnGroupBy:
         self.column = column
         
     def nunique(self):
-        return Column(self.gb.df.db, self.gb.df.table, sa.select(sa.func.count(self.gb.df.table.c[self.column].distinct())).group_by(self.gb.columns))
+        return Column(self.gb.df.db, sa.select(sa.func.count(self.gb.df.table.c[self.column].distinct())).group_by(*self.gb.columns))
     
     def count(self):
         return self.agg('count')
@@ -282,9 +290,9 @@ class ColumnGroupBy:
             columns.append(column)
 
         if len(columns) == 1:
-            return Column(self.gb.df.db, self.gb.df.table, sa.select(*columns).group_by(self.gb.columns))
+            return Column(self.gb.df.db, sa.select(*columns).group_by(*self.gb.columns))
         
-        return DataFrame(self.gb.df.db, self.gb.df.table, sa.select(*self.gb.df.table.c[tuple(self.gb.columns)], *columns).group_by(self.gb.columns))
+        return DataFrame(self.gb.df.db, sa.select(*self.gb.df.table.c[self.gb.columns], *columns).group_by(*self.gb.columns))
     
     def transform(self, *args, **kwargs):
         columns = []
@@ -300,16 +308,8 @@ class ColumnGroupBy:
             
         result_class = Column if len(columns) == 1 else DataFrame
         
-        return result_class(self.gb.df.db, self.gb.df.table, sa.select(*columns).order_by(sa.func.row_number().over()))
+        return result_class(self.gb.df.db, sa.select(*columns).order_by(sa.func.row_number().over()))
     
-    
-def decorator(func):
-    @functools.wraps(func)
-    def wrapped(self, *args):
-        args = [arg.column if isinstance(arg, Column) else arg for arg in args]
-        expr = func(self.column, *args)
-        return Column(self.db, expr, sa.select(expr))
-    return wrapped
 
 class Column(abc.SQL):
     def fetch(self):
@@ -317,17 +317,31 @@ class Column(abc.SQL):
             rows = con.execute(self.sql).fetchall()
         return np.array(rows).squeeze()
     
+    
+def decorator(func):
+    @functools.wraps(func)
+    def wrapped(self, *args):
+        args = [arg.column if isinstance(arg, IndexableColumn) else arg for arg in args]
+        expr = func(self.column, *args)
+        return IndexableColumn(self.db, expr)
+    return wrapped
+    
+
+class IndexableColumn(Column):
+    def __init__(self, db, column):
+        self.db = db
+        self.column = column
+    
     @property
     def str(self):
-        return StringMethods(self.db, self.column, self.sql)
+        return StringMethods(self.db, self.column)
     
     @property
-    def column(self):
-        return self.from_
+    def sql(self):
+        return sa.select(self.column)
     
-    @decorator
     def unique(self):
-        return self.distinct()
+        return Column(self.db, sa.select(self.column.distinct()))
     
     @decorator
     def __add__(self, x):
@@ -422,7 +436,7 @@ class Column(abc.SQL):
         return ~self
     
 
-class StringMethods(Column):
+class StringMethods(IndexableColumn):
     @decorator
     def contains(self, pat):
         return self.like(f'%{pat}%')
@@ -435,7 +449,7 @@ class StringMethods(Column):
     def endswith(self, pat):
         return self.like(f'%{pat}')
     
-                
+      
 class Row(abc.SQL):
     def fetch(self):
         with self.db.connect() as con:
